Require Import ZArith.
Require Import List.
Require Import util.
Require Import machine_int.
Import MachineInt.
Require Import state.
Require Import mips.
Require Import mips_tactics.

Open Local Scope asm_heap_scope.
Open Local Scope mips_scope.
Open Local Scope Z_scope.

Lemma singleton_equal: forall s s' m h1 h2 e1 e2 e3 e4,
  (e1 |-> e2) s s' m h1 ->
  (e3 |-> e4) s s' m h2 ->
  eval e1 s = eval e3 s ->
  eval e2 s = eval e4 s ->
  h1 = h2.
  intros.
  red in H.
  red in H0.
  inversion_clear H.
  inversion_clear H0.
  inversion_clear H3.
  inversion_clear H.
  rewrite H1 in H0.
  rewrite H0 in H3.
  assert (x = x0).
  omega.
  subst x0.
  rewrite H2 in H4.
  rewrite H4; rewrite H5; auto.
Qed.

Lemma cell_read: forall e v Q s s' m h,
 ((e |-> v) ** TT) s s' m h /\ Q s s' m h ->
 ((e |-> v) ** ((e |-> v) -* Q)) s s' m h.
  intros.
  inversion_clear H.
  Decompose_sepcon H0 h1 h2.
  Compose_sepcon h1 h2; auto.
  Intro_sepimp h1' h'.
  assert (h1' = h1).
  eapply singleton_equal.
  apply H4.
  apply H2.
  auto.
  auto.
  subst h1'.
  assert (h' = h).
  Equal_heap.
  rewrite H7; auto.
Qed.

(***************************************************************************)
(*                     Derived Hoare triples                               *)
(***************************************************************************)

(* addi *)

Lemma semax_addi' : forall P Q x y z,
  (P ==> update_store_addi x y z Q) ->
  {{ P }} addi x y z {{ Q }}.
  intros.
  apply semax_strengthen_pre with (update_store_addi x y z Q) .
  auto.
  apply semax_addi.
Qed.

Lemma semax_addi'' : forall R P Q x y z c,
  (P ==> update_store_addi x y z R) ->
  {{ R }} c {{ Q }} ->
  {{ P }} addi x y z ; c {{ Q }}.
  intros.
  apply semax_seq with R; try assumption.
  apply semax_addi'; assumption.
Qed.

(* addiu *)

Lemma semax_addiu' : forall P Q x y z,
  (P ==> update_store_addiu x y z Q) ->
  {{ P }} addiu x y z {{ Q }}.
  intros.
  apply semax_strengthen_pre with (update_store_addiu x y z Q) .
  auto.
  apply semax_addiu.
Qed.

Lemma semax_addiu'' : forall R P Q x y z c,
  (P ==> update_store_addiu x y z R) ->
  {{ R }} c {{ Q }} ->
  {{ P }} addiu x y z ; c {{ Q }}.
  intros.
  apply semax_seq with R; try assumption.
  apply semax_addiu'; assumption.
Qed.

(* addu *)

Lemma semax_addu' : forall P Q x y z,
  (P ==> update_store_addu x y z Q) ->
  {{ P }} addu x y z {{ Q }}.
intros.
  apply semax_strengthen_pre with (update_store_addu x y z Q) .
  auto.
  apply semax_addu.
Qed.

Lemma semax_addu'' : forall R P Q x y z c,
  (P ==> update_store_addu x y z R) ->
  {{ R }} c {{ Q }} ->
  {{ P }} addu x y z ; c {{ Q }}.
  intros.
  apply semax_seq with R; try assumption.
  apply semax_addu'; assumption.
Qed.

(* lw *)

Definition update_store_lw rt v P : assert :=
  fun s s' m h => P (gpr.update rt v s) s' m h.

Lemma semax_lw_backwards : forall rt index base Q,
  {{ fun s s' m h => exists e,
    (add_e (var_e base) (int_e (sign_extend_16_32 index)) |-> int_e e **
      ((add_e (var_e base) (int_e (sign_extend_16_32 index)) |-> int_e e) -* 
	(update_store_lw rt e Q))) s s' m h }} lw rt index base {{ Q }}.
  intros.
  apply semax_strengthen_pre with (lw2 rt index base Q).
  do 5 intro.
  inversion_clear H.
  generalize (sep.mp _ _ _ _ _ _ H0); intro.
  red in H.
  red.
  inversion_clear H0.
  inversion_clear H1.
  inversion_clear H0.
  inversion_clear H2.
  inversion_clear H3.
  inversion_clear H2 as [p].
  inversion_clear H3.
  exists p.
  split; auto.
  exists (eval (int_e x) s).
  split; auto.
  rewrite H0.
  apply heap.lookup_union; auto.
  rewrite H5.
  rewrite heap.lookup_singleton; auto.
  apply semax_lw.
Qed.

Lemma semax_lw_backwards'' : forall rt index base P Q R c,
  P ==> (fun s s' m h => exists e,
    (add_e (var_e base) (int_e (sign_extend_16_32 index)) |-> int_e e **
      ((add_e (var_e base) (int_e (sign_extend_16_32 index)) |-> int_e e) -* 
	(update_store_lw rt e R))) s s' m h) ->
  {{ R }} c {{ Q }} ->
  {{ P }} lw rt index base; c {{ Q }}.
  intros.
  apply semax_seq with (Q:=R).
  eapply semax_strengthen_pre.
  apply H.
  apply semax_lw_backwards.
  assumption.
Qed.

Lemma semax_lw_backwards''2 : forall rt index base P Q R c,
  P ==> (fun s s' m h => exists e,
    (add_e (var_e base) (int_e (sign_extend_16_32 index)) |-> int_e e ** TT) s s' m h /\      
    (update_store_lw rt e R) s s' m h) ->
  {{ R }} c {{ Q }} ->
  {{ P }} lw rt index base; c {{ Q }}.
  intros.
  apply semax_lw_backwards'' with (R:=R).
  red; intros.
  generalize (H _ _ _ _ H1); intros.
  inversion_clear H2.
  exists x.
  apply cell_read.
  assumption.
  auto.
Qed.

(* lwxs *)

Lemma semax_lwxs'' : forall R P Q rt index base c,
  (P ==> lookup_heap_lwxs rt index base R) ->
  {{ R }} c {{ Q }} ->
  {{ P }} lwxs rt index base ; c {{ Q }}.
  intros.
  apply semax_seq with (Q:=R).
  eapply semax_strengthen_pre.
  apply H.
  apply semax_lwxs.
  assumption.
Qed.

Definition update_store_lwxs rt v P : assert :=
  fun s s' m h => P (gpr.update rt v s) s' m h.

Lemma semax_lwxs_backwards : forall rt index base P,
  {{ fun s s' m h => exists e, 
    (add_e (var_e base) (shl2_e (var_e index)) |-> int_e e ** 
      ((add_e (var_e base) (shl2_e (var_e index)) |-> int_e e) -* 
	(update_store_lwxs rt e P))) s s' m h }}
  lwxs rt index base
  {{ P }}.
  intros.
  apply semax_strengthen_pre with (lookup_heap_lwxs rt index base P).
  do 5 intro.
  inversion_clear H.
  generalize (sep.mp _ _ _ _ _ _ H0); intro.
  red in H.
  red.
  inversion_clear H0.
  inversion_clear H1.
  inversion_clear H0.
  inversion_clear H2.
  inversion_clear H3.
  inversion_clear H2 as [p].
  inversion_clear H3.
  exists p.
  split; auto.
  exists (eval (int_e x) s).
  split; auto.
  rewrite H0.
  apply heap.lookup_union; auto.
  rewrite H5.
  rewrite heap.lookup_singleton; auto.
  apply semax_lwxs.
Qed.

Lemma semax_lwxs_backwards' : forall rt index base P Q,
  P ==> (fun s s' m h => exists e,
    (add_e (var_e base) (shl2_e (var_e index)) |-> int_e e **
      ((add_e (var_e base) (shl2_e (var_e index)) |-> int_e e) -* 
	(update_store_lwxs rt e Q))) s s' m h) ->
  {{P}} lwxs rt index base {{ Q }}.
  intros.
  eapply semax_strengthen_pre.
  apply H.
  apply semax_lwxs_backwards.
Qed.

Lemma semax_lwxs_backwards'' : forall rt index base P Q R c,
  P ==> (fun s s' m h => exists e,
    (add_e (var_e base) (shl2_e (var_e index)) |-> int_e e **
      ((add_e (var_e base) (shl2_e (var_e index)) |-> int_e e) -* 
	(update_store_lwxs rt e R))) s s' m h) ->
  {{ R }} c {{ Q }} ->
  {{ P }} lwxs rt index base; c {{ Q }}.
  intros.
  apply semax_seq with (Q:=R).
  eapply semax_strengthen_pre.
  apply H.
  apply semax_lwxs_backwards.
  assumption.
Qed.

Lemma semax_lwxs_backwards'2: forall rt index base P Q,
  (P ==> 
    (fun s s' m h => exists e,
      ((add_e (var_e base) (shl2_e (var_e index)) |-> int_e e) ** TT) s s' m h /\
      (update_store_lwxs rt e Q) s s' m h)) ->
  {{P}} lwxs rt index base {{ Q }}.
  intros.
  eapply semax_strengthen_pre; [idtac | apply semax_lwxs].
  red; intros.
  generalize (H _ _ _ _ H0); intros.
  inversion_clear H1.
  inversion_clear H2.
  red.
  Decompose_sepcon H1 h1 h2.
  red in H4.
  inversion_clear H4.
  inversion_clear H1.
  exists x0; split; auto.
  exists x.
  split.
  rewrite H5.
  apply heap.lookup_union.
  auto.
  simpl in H6.
  rewrite H6.
  rewrite heap.lookup_singleton; auto.
  red in H3; auto.
Qed.

Lemma semax_lwxs_backwards''2: forall rt index base P Q R c,
  (P ==> 
    (fun s s' m h => exists e,
      ((add_e (var_e base) (shl2_e (var_e index)) |-> int_e e) ** TT) s s' m h /\
      (update_store_lwxs rt e R) s s' m h)) ->
  {{R}} c {{Q}} ->
  {{P}} lwxs rt index base; c {{ Q }}.
  intros.
  apply semax_seq with R.
  apply semax_lwxs_backwards'2; auto.
  auto.
Qed.

(* maddu *)

Lemma semax_maddu' : forall (P Q:assert) rs rt,
  (forall s1 s2 m h, P s1 s2 m h -> 
    Q s1 s2 (multiplier.maddu_op (umul (gpr.lookup rs s1) (gpr.lookup rt s1)) m) h) ->
  {{ P }} maddu rs rt {{ Q }}.
  intros.
  apply semax_strengthen_pre with (fun s1 s2 m h => Q s1 s2 (multiplier.maddu_op (umul (gpr.lookup rs s1) (gpr.lookup rt s1)) m) h).
  auto.
  apply semax_maddu.
Qed.

Lemma semax_maddu'' : forall (P Q R:assert) rs rt c,
  (forall s s' m h, 
    P s s' m h -> R s s' (multiplier.maddu_op (umul (gpr.lookup rs s) (gpr.lookup rt s)) m) h) ->
  {{ R }} c {{ Q }} ->
  {{ P }} maddu rs rt ; c {{ Q }}.
  intros.
  apply semax_seq with R; try assumption.
  apply semax_maddu'; assumption.
Qed.
  
(* multu *)

Lemma semax_multu' : forall (P Q:assert) rs rt,
  (forall s s' m h, 
    P s s' m h -> Q s s' (multiplier.multu_op (umul (gpr.lookup rs s) (gpr.lookup rt s)) m) h) ->
  {{ P }} multu rs rt {{ Q }}.
  intros.
  apply semax_strengthen_pre with (fun s s' m h =>  Q s s' (multiplier.multu_op (umul (gpr.lookup rs s) (gpr.lookup rt s)) m) h).
  assumption.
  apply semax_multu.
Qed.

Lemma semax_multu'' : forall (P Q R:assert) rs rt c,
  (forall s s' m h, 
    P s s' m h -> R s s' (multiplier.multu_op (umul (gpr.lookup rs s) (gpr.lookup rt s)) m) h) ->
  {{ R }} c {{ Q }} ->
  {{ P }} multu rs rt ; c {{ Q }}.
  intros.
  eapply semax_seq.
  eapply semax_multu'.
  apply H.
  auto.
Qed.

(* mflhxu *)

Lemma semax_mflhxu' : forall P Q x,
  (P ==> update_store_mflhxu x Q) ->
  {{ P }} mflhxu x {{ Q }}.
  intros.
  eapply semax_strengthen_pre.
  apply H.
  apply semax_mflhxu.
Qed.

Lemma semax_mflhxu'' : forall R P Q x c,
  (P ==> update_store_mflhxu x R) ->
  {{ R }} c {{ Q }} ->
  {{ P }} mflhxu x ; c {{ Q }}.
  intros.
  eapply semax_seq.
  eapply semax_mflhxu'.
  apply H.
  auto.
Qed.

(* mflo *)

Lemma semax_mflo' : forall (P Q:assert) rd,
  P ==> update_store_mflo rd Q ->
  {{ P }} mflo rd {{ Q }}.
  intros.
  eapply semax_strengthen_pre.
  apply H.
  apply semax_mflo.
Qed.

Lemma semax_mflo'' : forall (P Q R:assert) rd c,
  P ==> update_store_mflo rd R ->
  {{ R }} c {{ Q }} ->
  {{ P }} mflo rd ; c {{ Q }}.
  intros.
  eapply semax_seq.
  eapply semax_mflo'.
  apply H.
  auto.
Qed.

(* mfhi *)

Lemma semax_mfhi' : forall (P Q:assert) rd,
  P ==> update_store_mfhi rd Q ->
  {{ P }} mfhi rd {{ Q }}.
  intros.
  eapply semax_strengthen_pre.
  apply H.
  apply semax_mfhi.
Qed.

Lemma semax_mfhi'' : forall (P Q R:assert) rd c,
  P ==> update_store_mfhi rd R ->
  {{ R }} c {{ Q }} ->
  {{ P }} mfhi rd ; c {{ Q }}.
  intros.
  eapply semax_seq.
  eapply semax_mfhi'.
  apply H.
  auto.
Qed.

(* mthi *)

Lemma semax_mthi' : forall (P Q:assert) rs,
  P ==> update_store_mthi rs Q ->
  {{ P }} mthi rs {{ Q }}.
  intros.
  eapply semax_strengthen_pre.
  apply H.
  apply semax_mthi.
Qed.

Lemma semax_mthi'' : forall (P Q R:assert) rs c,
  P ==> update_store_mthi rs R ->
  {{ R }} c {{ Q }} ->
  {{ P }} mthi rs ; c {{ Q }}.
  intros.
  eapply semax_seq.
  eapply semax_mthi'.
  apply H.
  auto.
Qed.

(* mtlo *)

Lemma semax_mtlo' : forall (P Q:assert) rs,
  P ==> update_store_mtlo rs Q ->
  {{ P }} mtlo rs {{ Q }}.
  intros.
  eapply semax_strengthen_pre.
  apply H.
  apply semax_mtlo.
Qed.

Lemma semax_mtlo'' : forall (P Q R:assert) rs c,
  P ==> update_store_mtlo rs R ->
  {{ R }} c {{ Q }} ->
  {{ P }} mtlo rs ; c {{ Q }}.
  intros.
  eapply semax_seq.
  eapply semax_mtlo'.
  apply H.
  auto.
Qed.

(* sw *)

Lemma semax_sw' : forall P Q rt offset base,
  (P ==> update_heap2 rt offset base Q) ->
  {{ P }} sw rt offset base {{ Q }}.
  intros.
  eapply semax_strengthen_pre.
  apply H.
  apply semax_sw.
Qed.

Lemma semax_sw'' : forall R P Q rt offset base c,
  (P ==> update_heap2 rt offset base R) ->
  {{ R }} c {{ Q }} ->
  {{ P }} sw rt offset base ; c {{ Q }}.
  intros.
  apply semax_seq with R; try assumption.
  apply semax_sw'; assumption.
Qed.

Lemma semax_sw_global : forall P rt idx base,
  {{ (fun s1' s2' m' h' => 
    exists e'', ((add_e (var_e base) (int_e (sign_extend_16_32 idx)) |-> e'') s1' s2' m' h')) ** P}} 
  sw rt idx base 
  {{ (fun s1 s2 m h => (add_e (var_e base) (int_e (sign_extend_16_32 idx)) |-> 
    var_e rt) s1 s2 m h) ** P }}.
  intros.
  apply frame_rule.
  apply semax_strengthen_pre with (update_heap2 
    rt idx base (fun s1 s2 m h => 
      (add_e (var_e base) (int_e (sign_extend_16_32 idx)) |-> var_e rt) s1 s2 m h)).
  red; intros.
  inversion_clear H.
  red in H0.
  inversion_clear H0.
  red.
  exists x0; split; intuition.
  exists (eval x s).
  rewrite H1.
  apply heap.lookup_singleton.
  red.
  exists x0; split; auto.
  simpl eval.
  apply trans_eq with (heap.update x0 (gpr.lookup rt s) (heap.singleton x0 (eval x s))).
  rewrite H1.
  auto.
  apply heap.update_singleton.
  apply semax_sw.
  simpl.
  red; intros.
  simpl in H; contradiction.
  red.
  intros.
  simpl in H.
  tauto.
Qed.

Lemma semax_sw_backwards : forall rt idx base Q,
  {{ fun s s' m h => exists e,
    (add_e (var_e base) (int_e (sign_extend_16_32 idx)) |-> e **
      ((add_e (var_e base) (int_e (sign_extend_16_32 idx)) |-> var_e rt) -* Q)) s s' m h }}
  sw rt idx base
  {{ Q }}.
  intros.
  apply semax_strengthen_pre with (update_heap2 rt idx base Q).
  red; intros.
  inversion_clear H.
  inversion_clear H0 as [h1].
  inversion_clear H as [h2].
  decompose [and] H0; clear H0.
  red.
  inversion_clear H1 as [p].
  inversion_clear H0.
  exists p.
  split.
  auto.
  split.
  exists (eval x s).
  rewrite H2.
  apply heap.lookup_union.
  auto.
  rewrite H3.
  apply heap.lookup_singleton.
  apply H4 with (heap.update p (gpr.lookup rt s) h1).
  split.
  apply heap.disjoint_com.
  apply heap.disjoint_update.
  auto.
  exists p.
  split; auto.
  simpl.
  rewrite H3.
  apply heap.update_singleton.
  rewrite H2.
  rewrite heap.union_com.
  eapply heap.equal_update_R.
  apply heap.disjoint_com; auto.
  rewrite H3.
  apply heap.lookup_singleton.
  auto.
  apply semax_sw.
Qed.
(* an alternative proof using the frame rule:
intros.
generalize (semax_sw_global 
  (add_e (var_e base) (int_e (sign_extend_16_32 idx))
         |-> var_e rt -* Q) rt idx base); intro.
apply semax_weaken_post with (
  ((add_e (var_e base) (int_e (sign_extend_16_32 idx))
         |-> var_e rt) **
  (add_e (var_e base) (int_e (sign_extend_16_32 idx))
         |-> var_e rt -* Q))
  ).
apply sep.mp.
apply semax_strengthen_pre with (
(fun s1 s2 m h => exists e'', (add_e (var_e base) (int_e (sign_extend_16_32 idx))
         |-> e'') s1 s2 m h) ** (add_e (var_e base) (int_e (sign_extend_16_32 idx))
         |-> var_e rt -* Q)).
red; intros.
inversion_clear H0.
red.
inversion_clear H1 as [h1].
inversion_clear H0 as [h2].
decompose [and] H1; clear H1.
exists h1; exists h2.
split; auto.
split; auto.
split.
exists x.
auto.
auto.
auto.
Qed.
*)

Lemma semax_sw_backwards' : forall rt index base P Q,
  P ==> (fun s s' m h => exists e,
    (add_e (var_e base) (int_e (sign_extend_16_32 index)) |-> e **
      ((add_e (var_e base) (int_e (sign_extend_16_32 index)) |-> var_e rt) -* Q)) s s' m h) ->
  {{P}} sw rt index base {{ Q }}.
  intros.
  eapply semax_strengthen_pre.
  apply H.
  apply semax_sw_backwards.
Qed.

Lemma semax_sw_backwards'' : forall rt index base P Q R c,
  P ==> (fun s s' m h => exists e,
    (add_e (var_e base) (int_e (sign_extend_16_32 index)) |-> e **
      ((add_e (var_e base) (int_e (sign_extend_16_32 index)) |-> var_e rt) -* R)) s s' m h) ->
  {{ R }} c {{ Q}} ->
  {{ P }} sw rt index base ; c {{ Q }}.
  intros.
  eapply semax_seq.
  apply semax_sw_backwards'.
  apply H.
  auto.
Qed.

Lemma semax_assign_lookup_expr_global : forall P rt idx base,
  {{ (fun s s' m h => exists e, 
    (add_e (var_e base) (int_e (sign_extend_16_32 idx)) |-> e ) s s' m h) ** P }} 
  sw rt idx base
  {{(add_e (var_e base) (int_e (sign_extend_16_32 idx)) |-> var_e rt ) ** P }}.
  intros.
  apply frame_rule.
  apply semax_strengthen_pre with (update_heap2 rt idx base (
    (add_e (var_e base) (int_e (sign_extend_16_32 idx)) |-> var_e rt ))).
  red; intros.
  inversion_clear H.
  red in H0.
  inversion_clear H0.
  red.
  exists x0; split; intuition.
  exists (eval x s).
  rewrite H1.
  rewrite heap.lookup_singleton.
  reflexivity.
  exists x0; split.
  auto.
  simpl.
  rewrite H1.
  rewrite heap.update_singleton.
  reflexivity.
  apply semax_sw.
  simpl.
  red; intros.
  simpl in H; tauto.
  red; intros.
  simpl in H; tauto.
Qed.

(* sltu *)

Lemma semax_sltu'' : forall R P Q rd rs rt c,
  (P ==> update_store_sltu rd rs rt R) ->
  {{ R }} c {{ Q }} ->
  {{ P }} sltu rd rs rt ; c {{ Q }}.
  intros.
  apply semax_seq with R; auto.
  eapply semax_strengthen_pre; [idtac | apply semax_sltu].
  auto.
Qed.  

(* msubu *)  

Lemma semax_msubu'' : forall (P Q R:assert) rs rt c,
  (forall s s' m h, 
    P s s' m h -> R s s' (multiplier.msubu_op (umul (gpr.lookup rs s) (gpr.lookup rt s)) m) h) ->
  {{ R }} c {{ Q }} ->
  {{ P }} msubu rs rt ; c {{ Q }}.
  intros.
  apply semax_seq with R; auto.
  eapply semax_strengthen_pre; [idtac | constructor].
  auto.
Qed.  






